#include <Arduino.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <AsyncElegantOTA.h>

#include <TM1638plus.h>
#include <SPIFFS.h>
#include <Wire.h>

#include "logic.h"
#include "function_map.h"
#include "main.h"

/*#define BLOCK_AMOUNT 100
#define PIN_AMOUNT 300
#define OUTPUT_NR 10*/

#define BUT8 0b10000000
#define BUT7 0b01000000
#define BUT6 0b00100000
#define BUT5 0b00010000
#define BUT2 0b00000010
#define BUT1 0b00000001

#define STR_LEN 10
#define FILENAME_LEN 7

#define AUX_PORT_WIDTH 8
#define AUX_PORT_AMOUNT 2

#define AUX_MAX_DEV 10

#define SLAVE_ADDR 0x42
#define END_CHAR 0xff
#define EMPTY_CHAR 0xff

#define I2C_SPEED 400000

#define AUX_UPDATE_INTERVAL 100

#define SIM_TIME_LED 7
#define SLOW_SIM_LED 6
#define SLOW_SIM_DELAY 1000

TM1638plus disp(TM1638_STROBE, TM1638_CLOCK, TM1638_DATA, true);

const char* ssid = "ssid";
const char* password = "password";

AsyncWebServer server(80);

//pole vsech funkcnich bloku
block blocks[BLOCK_AMOUNT];

//pole vsech pinu
pin pins[PIN_AMOUNT];

//pole vsech dratu
wire wires[WIRE_AMOUNT];
unsigned int wires_used = 1;

//globalni pole pro vstupy a vystupy dratu
pin_number wire_inputs[WIRE_AMOUNT * INPUT_NR];
pin_number wire_outputs[WIRE_AMOUNT * INPUT_NR];

//vychozi globalni pole pro vstupy a vystupy dratu
//naplni se hodnotami ze souboru a pak slouzi jako vychozi pri testovani fyzickych propojek
//-> pred kazdou obnovou stavu propojek se pouzivane pole prepisou hodnotou z vychozich
pin_number default_wire_inputs[WIRE_AMOUNT * INPUT_NR];
pin_number default_wire_outputs[WIRE_AMOUNT * INPUT_NR];

//globalni pole cisel dratu, na kterem jsou pripojeny piny
wire_number pins_wires[PIN_AMOUNT];
//vychozi globalni pole cisel dratu, na kterem jsou pripojeny piny (stejny duvod jako pole vyse)
wire_number default_pins_wires[PIN_AMOUNT];


block_number compute_order[BLOCK_AMOUNT];
unsigned char start_block_index = 0;

unsigned long disp_button_debounce;
unsigned char last_buttons = 0;

signed char device_conf_nr = 0;
signed char connection_conf_nr = -1;

char disp_err = NO_ERR;
unsigned char err_num1 = 0;
unsigned char err_num2 = 0;

char aux_devices = 0;

char sim_time_enable = TIME_OFF;
//inicializace tak, aby pri prvnim pruchodu vzdy doslo k nacteni fyzickych propojek
unsigned long sim_time = AUX_UPDATE_INTERVAL + 1;

char slow_sim_enable = TIME_OFF;
unsigned long slow_last_run = 0;

//matice uchovavajici stav fyzickych propojek
unsigned char aux_matrix[AUX_PORT_AMOUNT * AUX_PORT_WIDTH][AUX_PORT_AMOUNT * AUX_PORT_WIDTH / 8];

unsigned char last_aux_update = 0;

/**
 * @brief Obnovi vypisovana cisla na 7 segmentovych displejich
 * 
 */
void display_refresh() {
	//LED zhasnu, v pripade potreby se znovu rozsviti
	disp.setLEDs(0);

	if(slow_sim_enable == TIME_ON) {
		disp.setLED(SLOW_SIM_LED, 1);
	}
	if(sim_time_enable == TIME_ON) {
		disp.setLED(SIM_TIME_LED, 1);
	}

	disp.displayHex(4, device_conf_nr / 10); //konfigurace soucastek
	disp.displayHex(5, device_conf_nr % 10);
	
	if(connection_conf_nr < 0) {
		disp.display7Seg(6, 0b01000000); //dve pomlcky jako cislo konfigurace propojek (= nepouzito)
		disp.display7Seg(7, 0b01000000);
	}
	else {
		disp.displayHex(6, connection_conf_nr / 10); //konfigurace soucastek
		disp.displayHex(7, connection_conf_nr % 10);
	}

	if(disp_err != NO_ERR) {
		#ifdef PRINT
		Serial.printf("disp_err: %d\n", disp_err);
		#endif

		disp.setLED(disp_err - 1, 1);
		//pri invalid connection a runtime error je v err_num2 cislo pinu, muze byt az triciferne
		if(disp_err == RUN_ERR || disp_err == INV_CONN_ERR || disp_err == PHY_CONN_ERR) {
			disp.displayHex(1, err_num2 / 100);
			disp.displayHex(2, (err_num2 % 100) / 10);
			disp.displayHex(3, err_num2 % 10);
		}
		//pri errorech z nacitani souboru jsou cisla dve, maximalne dvouciferna
		else {
			disp.displayHex(0, err_num1 / 10);
			disp.displayHex(1, err_num1 % 10);
			disp.displayHex(2, err_num2 / 10);
			disp.displayHex(3, err_num2 % 10);
		}
	}
	else { //pokud neni zadna chyba, prvni ctyri sedmisegmenty zhasnu
		disp.display7Seg(0, 0);
		disp.display7Seg(1, 0);
		disp.display7Seg(2, 0);
		disp.display7Seg(3, 0);
	}
}

void show_error(char type, unsigned char line, unsigned char object) {
	err_num1 = line;
	err_num2 = object;
	disp_err = type;
	display_refresh();
}

/**
 * @brief Nacte ze souboru specifikovanem argumentem num soucastky a ulozi je do globalnich prommenych pins a blocks
 * 
 * @param num cislo souboru
 * @return char 0 v pripade uspechu, >0 pokud doslo k chybe
 */
char get_devices(char num) {
	char n[FILENAME_LEN];
	memset(n, 0, FILENAME_LEN * sizeof(char));

	n[0] = '/';
	n[1] = num / 10 + '0';
	n[2] = num % 10 + '0';
	n[3] = '.';
	n[4] = 'P';
	n[5] = 'C';
	n[6] = 0;

	char s[STR_LEN];
	memset(s, 0, STR_LEN * sizeof(char));
	unsigned char s_index = 0;
	//String s;

	char r;
	pin_number pin_buf;
	unsigned char to_read = 0;

	unsigned char input_index = 0;
	unsigned char output_index = 0;

	unsigned char line = 1;
	unsigned char device = 1;

	unsigned char success = 0;

	//znulovani chyby
	err_num1 = 0;
	err_num2 = 0;

	//block 0 bude nevyuzity pro snazsi zastavovani cyklu
	int i = 1;
	int j = 0;
	int k = 0;

	memset(blocks, 0, sizeof(block) * BLOCK_AMOUNT);
	memset(pins, 0, sizeof(pin) * PIN_AMOUNT);

	memset(pins_wires, 0, PIN_AMOUNT * sizeof(wire_number));
	//pole pins jsem cele znuloval, musim znovu zapsat ukazatele
	for(int i = 0; i < PIN_AMOUNT; ++i) {
		pins[i].wire_nr = pins_wires + i;
	}

	memset(compute_order, 0, sizeof(compute_order));
	start_block_index = 0;

	Serial.printf("Filename: %s\n", n);
	if(!SPIFFS.exists(n)) {
		return 1;
	}
	File file = SPIFFS.open(n, "r");

	if (!file) {
		Serial.println("Open ERR");
		return 1;
	}

	while (file.available()) {
		r = file.read();

		//mezi soucastkami muze byt odradkovani
		if(r == '\n') {
			//pokud neni mezi soucastkami, bere se jako chyba
			/*if(s_index != 0) {
				show_error(DEV_ERR, line, device);
				printf("%s  %d\n", s, s_index);
				file.close();
				return 1;
			}*/
			device = 0;
			++line;
			continue;
		}

		//nacitam dokud nanarazim na pomlcku
		if (r != '-') {
			s[s_index] = r;
			++s_index;
		}

		//mam nacteny typ bloku
		else {
			success = 0;
			k = 0;
			//mapovaci tabulka je ukoncena nulovym prvkem
			while(mapping[k].name != NULL) {
				//pokud najdu polozku, ktera se shoduje s tim, co jsem nacetl
				if(!strncmp(mapping[k].name, s, s_index)) {
					success = 1;
					++device;

					blocks[i].function_pointer = mapping[k].function_pointer;
					blocks[i].inputs_used = mapping[k].input_amount;
					to_read = mapping[k].input_amount;

					#ifdef PRINT
					Serial.printf("to_read: %d\n", (int)to_read);
					#endif

					if(mapping[k].starting == STARTING) {
						compute_order[start_block_index] = i;
						++start_block_index;
					}

					#ifdef PRINT
					Serial.printf("Nacten %s.\n", mapping[k].name);
					#endif

					break;
				}
				++k;
			}
			if(!success) {
				show_error(DEV_ERR, line, device);
				file.close();
				return 1;
			}
			
			s_index = 0;
			pin_buf = 0;		
			j = 0;

			while(1) {
				
				//pokud by soucastka nebyla korektne ukoncena, je potreba nacitani nejak zastavit a informovat uzivatele
				if(!file.available()) {
					show_error(DEV_ERR, line, device);
					file.close();
					return 1;
				}

				r = file.read();
		
				if (r == '-' || r == ';') {
					//jedno cislo pinu jsem docetl (oddeleni pomlckami)
					pins[pin_buf].io_nr = j;
					++j;
					
					if (to_read > 0) {
						//nacitane piny jsou vstupni, io_nr ma MSB == 0
						--to_read;

						//do bloku zapisu cislo vstupniho pinu
						blocks[i].input[input_index] = pin_buf;
						++input_index;

						if(!to_read) {
							//docetl jsem vstupy, j bude ted udavat indexy vystupu
							j = 0;
						}
					}
					else {
						//pokud by mela soucastka vic vystupu nez je velikost vyhrazeneho pole
						if(output_index >= OUTPUT_NR) {
							show_error(DEV_ERR, line, device);
							file.close();
							return 1;
						}
						//pin je vystupni, io_nr bude mit MSB == 1
						pins[pin_buf].io_nr = pins[pin_buf].io_nr | IO_OUTPUT;

						//do bloku zapisu cislo vystupniho pinu
						blocks[i].output[output_index] = pin_buf;

						//pokud jsem nacetl buffer, jeho vystup musi priznak tristate
						if(!strcmp(mapping[k].name, "BUF")) {
							pins[pin_buf].tristate = TRISTATE_TRUE;
						}
						
						#ifdef PRINT
						Serial.printf("Pin %d je vystupni.\n", pin_buf);
						#endif
					}
					
					#ifdef PRINT
					Serial.printf("Zapsan pin %d\n", (int)pin_buf);
					#endif

					pins[pin_buf].block_nr = i;
					pin_buf = 0;

					if(r == ';') {
						//pokud je znak strednik, soucastka konci
						#ifdef PRINT
						Serial.println("Dokoncen zapis soucastky.");
						#endif

						break;
					}
				}
				else {
					if(r >= '0' && r <= '9') {
						pin_buf = pin_buf * 10 + (r - '0');
					}
					else {
						//pokud narazim na nejaky necekany znak, je to chyba
						show_error(DEV_ERR, line, device);
						file.close();
						return 1;
					}
				}
			}
			++i;
			input_index = 0;
			output_index = 0;
			//znulovani pro nacteni dalsiho jmena soucastky
			memset(s, 0, STR_LEN * sizeof(char));
		}
	}
	file.close();

	//pokud po konci nacitani zustaly nezpracovane znaky, nekde je chyba
	if(s_index != 0) {
		show_error(DEV_ERR, line, device);
		return 1;
	}
	return 0;
}

char bitmap_is_set(unsigned char *bitmap, unsigned char item_nr) {
	return bitmap[item_nr / 8] & (0b10000000 >> (item_nr % 8));
}

void bitmap_set(unsigned char *bitmap, unsigned char item_nr) {
	bitmap[item_nr / 8] = bitmap[item_nr / 8] | (0b10000000 >> (item_nr % 8));
}

/**
 * @brief Podle nactenych propojek sestavi poradi, v jakem se bloky budou vykonavat a ulozi ho do pole compute_order
 * 
 */
void get_block_order() {
	unsigned char wire_bitmap[(WIRE_AMOUNT / 8) + 1]; //pro evidenci dratu, ktere uz byly projity
	unsigned char block_bitmap[(BLOCK_AMOUNT / 8) + 1]; //pro evidenci bloku, ktere uz byly projity
	unsigned char wire_index = 1;
	unsigned char block_index = start_block_index;
	wire_number output_wire;
	block_number output;
	block_number block_nr;

	pin_number temp;
	block b;
	  
	memset(wire_bitmap, 0, sizeof(wire_bitmap));
	memset(block_bitmap, 0, sizeof(block_bitmap));
	
	//pole compute order znuluju az na bloky oznacene jako pocatecni
	memset(compute_order + start_block_index, 0, sizeof(compute_order) - start_block_index);

	#ifdef PRINT
	Serial.printf("Zjistuji poradi bloku.\n");
	#endif

	for(int i = 0; i < start_block_index; ++i) {
    	block_nr = compute_order[i];
    	//do bitmapy se zaznamenaji vstupni bloky (do pole compute_order byly zadany uz pri nacitani bloku ze souboru)
		bitmap_set(block_bitmap, block_nr);
    	//block_bitmap[block_nr / 8] = block_bitmap[block_nr / 8] | (0b10000000 >> (block_nr % 8));
  	}
		
	//postupne projdu vsechny bloky ve fronte a pridam jejich nasledniky podle pripojenych dratu
	for(int i = 0; i < BLOCK_AMOUNT; ++i) {
		
		if(i > block_index) {
			//pokud uz bych se s indexem dostal pres obsazenou cast pole
			break;
		}

		b = blocks[compute_order[i]];

		#ifdef PRINT
		Serial.printf("Prochazim blokem %d.\n", compute_order[i]);
		#endif

		for(int j = 0; j < OUTPUT_NR; ++j) {
			temp = b.output[j];

			output_wire = *pins[temp].wire_nr;
			//output_wire = wire_from_pin(blocks[i].output[j]);
			
			if(!output_wire) {
				//je mozne, ze na vystupu nic pripojeno neni nebo se doslo na konec pole
				break;
			}

			//pokud jsem dany drat v podle bitmapy jeste neprosel (ma hodnotu 0)
			//if(!(wire_bitmap[output_wire / 8] & (0b10000000 >> (output_wire % 8)))) {
			if(!bitmap_is_set(wire_bitmap, output_wire)) {
				//oznacim ho jako projity a pridam do fronty zpracovani vsechny bloky, ktere z neho cerpaji vstup
				bitmap_set(wire_bitmap, output_wire);
		        //wire_bitmap[output_wire / 8] = wire_bitmap[output_wire / 8] | (0b10000000 >> (output_wire % 8));

				for(int k = 0; k < OUTPUT_NR; ++k) {
					temp = wires[output_wire].inputs[k];
					output = pins[temp].block_nr;
					
					if(!output) {
						break;
					}
					
					//blok do bitmapy (a fronty) pridam jen tehdy, pokud tam jeste neni
					//if(!(block_bitmap[output / 8] & (0b10000000 >> (output % 8)))) {
					if(!bitmap_is_set(block_bitmap, output)) {
						bitmap_set(block_bitmap, output);
						//block_bitmap[output / 8] = block_bitmap[output / 8] | (0b10000000 >> (output % 8));
						
						compute_order[block_index] = output;
						++block_index;

						#ifdef PRINT
						Serial.printf("Pridavam do fronty blok %d\n", output);
						#endif
					}
				}			
		        ++wire_index;
		    }
		}
	}
	
	#ifdef PRINT
	Serial.printf("Poradi bloku: ");
	for(int i = 0; i < BLOCK_AMOUNT; ++i) {
	    if(!compute_order[i]) {
	    	break;
	    }
    	Serial.printf("%d ", compute_order[i]);
	}
	Serial.printf("\n");
	#endif
	
	return;
}


/**
 * @brief Spoji draty specifikovane jejich cisly, predane v poli w do jednoho dratu (prvniho v poli w)
 * 
 * @param w pole dratu, ktere se maji sloucit
 * @param w_len delka pole w
 * @return char 0 v pripade uspechu, >0 pokud doslo k chybe
 */
char merge_wires(wire_number *w, unsigned char w_len) {
	unsigned char input_index = 0;
	unsigned char output_index = 0;
	pin_number pin_nr = 0;
	
	//bitmapa bude slouzit jako prevence pred nasobnym vkladanim stejnych pinu na jeden drat
	unsigned char pin_bitmap[PIN_AMOUNT / 8];
	memset(pin_bitmap, 0, PIN_AMOUNT / 8);

	//vsechny piny musim smrsknout na jeden drat - treba na ten prvni v predanem poli
	//v poli wires tim bohuzel vzniknou diry nevyuzitych a nevyuzitelnych dratu
	wire_number src_wire = 0;
	wire_number dest_wire = w[0];
	
	#ifdef PRINT
	Serial.printf("Spojuji draty.\n");
	#endif

	//nejdrive vstupni piny
	
	while(wires[dest_wire].inputs[input_index]) {
		pin_nr = wires[dest_wire].inputs[input_index];
		//predpoklad, ze v puvodnim dratu zadne nasobne piny nejsou
		bitmap_set(pin_bitmap, pin_nr);
		//pin_bitmap[pin_nr / 8] = pin_bitmap[pin_nr / 8] | (0b10000000 >> pin_nr % 8);
		
		++input_index;
	}
		
	for(int i = 1; i < w_len; ++i) {
		src_wire = w[i];
		
		for(int j = 0; j < INPUT_NR; ++j) {
			pin_nr = wires[src_wire].inputs[j];
			
			if(!pin_nr) {
				break;
			}
			//pin zapisu jedine, pokud neni v bitmape
			//if(!(pin_bitmap[pin_nr / 8] & (0b10000000 >> pin_nr % 8))) {
			if(!bitmap_is_set(pin_bitmap, pin_nr)) {
				bitmap_set(pin_bitmap, pin_nr);
				wires[dest_wire].inputs[input_index] = wires[src_wire].inputs[j];
				++input_index;
			}
			//ale zmenit cislo dratu musim vzdy
			*pins[pin_nr].wire_nr = dest_wire;
		}
	}
	
	memset(pin_bitmap, 0, PIN_AMOUNT / 8);
	
	//vystupni piny - musim kontrolovat i aby mely tristate pokud jich je vic, jinak je to chyba zapojeni
	
	while(wires[dest_wire].outputs[output_index]) {
		pin_nr = wires[dest_wire].outputs[output_index];
		//predpoklad, ze v puvodnim dratu zadne nasobne piny nejsou
		//pin_bitmap[pin_nr / 8] = pin_bitmap[pin_nr / 8] | (0b10000000 >> pin_nr % 8);
		bitmap_set(pin_bitmap, pin_nr);
		
		++output_index;
	}
		
	for(int i = 1; i < w_len; ++i) {
		src_wire = w[i];
		
		for(int j = 0; j < OUTPUT_NR; ++j) {
			pin_nr = wires[src_wire].outputs[j];
			
			if(!pin_nr) {
				break;
			}
			
			//pin mohu zapsat pokud neni v bitmape
			//8if(!(pin_bitmap[pin_nr / 8] & (0b10000000 >> pin_nr % 8))) {
			if(!bitmap_is_set(pin_bitmap, pin_nr)) {
				bitmap_set(pin_bitmap, pin_nr);
				if(output_index) { //ma drat uz nejake vystupy?
					
					if(output_index == 1) { //drat ma prave jeden vystup, musim kontrolovat tristate na nem i na pridavanem
						if(!(pins[wires[dest_wire].outputs[output_index - 1]].tristate == TRISTATE_TRUE && pins[pin_nr].tristate == TRISTATE_TRUE)) {
							//jeden z pinu tristate nema
							

							//show_error(INV_CONN_ERR, 0, pin_nr);

							#ifdef PRINT
							Serial.printf("Chyba pri spojovani vodicu - vice vystupu nema tri stavy!\n");
							#endif
							
							return 1;
						}
					}
					else { //vystupu uz je vic jak dva, staci kontrolovat tristate na nove pridavanem
						if(pins[pin_nr].tristate == TRISTATE_FALSE) {
							//pridavany pin tristate nema

							//show_error(INV_CONN_ERR, 0, pin_nr);

							#ifdef PRINT
							Serial.printf("Chyba pri spojovani vodicu - vice vystupu nema tri stavy!\n");
							#endif
							
							return 1;
						}
					}
				}
				//pin pridam mezi vystupy dratu
				wires[dest_wire].outputs[output_index] = wires[src_wire].outputs[j];
				++output_index;
			}
			*pins[pin_nr].wire_nr = dest_wire;
		}
	}
	return 0;
}


void clear_wires() {
	memset(wires, 0, WIRE_AMOUNT * sizeof(wire));
	memset(wire_inputs, 0, WIRE_AMOUNT * sizeof(pin_number));
	memset(wire_outputs, 0, WIRE_AMOUNT * sizeof(pin_number));
  	//"rozkouskovani" pole vstupu a vystupu do konkretnich dratu
	for(int i = 0; i < WIRE_AMOUNT; ++i) {
		wires[i].inputs = wire_inputs + i * INPUT_NR;
		wires[i].outputs = wire_outputs + i * OUTPUT_NR;
	}

	wires[0].wire_state = STATE_UNDEF;

	memset(pins_wires, 0, PIN_AMOUNT * sizeof(wire_number));
	//"rozkouskovani" pole cisel dratu, ke kterym jsou pripojene piny
	for(int i = 0; i < PIN_AMOUNT; ++i) {
		pins[i].wire_nr = pins_wires + i;
	}
}

/**
 * @brief Prida novy drat, spojujici piny predane v poli pin_buf
 * Pokud by piny uz byly k nejakym jinym dratum pripojene, necha je spojit funkci merge_wires
 * 
 * @param pin_buf cisla pinu, ktere maji byt spojene novym dratem
 * @return char 0 pokud vse probehlo v poradku, 1 pokud doslo k chybe
 */
char add_wire(pin_number *pin_buf) {
	unsigned char input_index = 0;
	unsigned char output_index = 0;
	
	pin_number output_pin = 0;
	
	wire_number wired_already[INPUT_NR + OUTPUT_NR];
	unsigned char wired_index = 0;
	
	unsigned char result = 0;
	
	memset(wired_already, 0, INPUT_NR + OUTPUT_NR);

	#ifdef PRINT
	Serial.printf("Vytvarim drat c.%d.\n", wires_used);
	for(int i = 0; i < INPUT_NR + OUTPUT_NR; ++i) {
		Serial.printf("%d ", pin_buf[i]);
	}
	Serial.printf("\n");
	#endif
	
	//vytvorim pro mnozinu pinu novy drat a zapisu je na nej
	for (int i = 0; i < INPUT_NR + OUTPUT_NR; ++i) {
		if(!pin_buf[i]) {
			break;
		}
		if(*pins[pin_buf[i]].wire_nr) {
			//pin uz je pripojen na nejakem drate - bude potreba dale resit
			//kdyby nahodou byl jeden pin uveden na jednom propoji dvakrat, delalo by to potize - kontrola
			if(*pins[pin_buf[i]].wire_nr != wires_used) {
				#ifdef PRINT
				Serial.printf("Pin %d uz je pripojen na drate %d.\n", pin_buf[i], *pins[pin_buf[i]].wire_nr);
				#endif

				//neresi se, pokud je vic pinu ze stejneho jineho dratu - merge_wires si s tim poradi, i kdyz za cenu vice pruchodu cykly - TODO?
				wired_already[wired_index] = *pins[pin_buf[i]].wire_nr;
				++wired_index;
			}
		}
		//kontrola, jestli je pin podle pole pinu vystupni
		if (pins[pin_buf[i]].io_nr & IO_OUTPUT) {
			if(output_pin > 0) {
				//v zapojeni je vic jak jeden vystupni pin - maji tristavove vystupy?
				if(!(pins[output_pin].tristate == TRISTATE_TRUE && pins[pin_buf[i]].tristate == TRISTATE_TRUE)) {
					//pokud ne, je to chyba
					
					Serial.printf("Spojene vystupni piny nemaji tristavove vystupy.\n");
					return 1;
				}
			}
			output_pin = pin_buf[i];
			wires[wires_used].outputs[output_index] = pin_buf[i];
			++output_index;
		}
		else {
			wires[wires_used].inputs[input_index] = pin_buf[i];
			++input_index;
		}
		*pins[pin_buf[i]].wire_nr = wires_used;
	}
	
	if(wired_index) {
		wired_already[wired_index] = wires_used;
		++wired_index;
		result = merge_wires(wired_already, wired_index);
		//posledni pouzity drat jsem spojenim zlikvidoval a da se dale pouzit - jen znuluju jeho vstupy a vystupy
		memset(wires[wires_used].inputs, 0, INPUT_NR);
		memset(wires[wires_used].outputs, 0, OUTPUT_NR);
		--wires_used;
	}
	
	++wires_used;
	
	return result;
}

/**
 * @brief Ze souboru definovanem jeho cislem nacte propojky mezi bloky
 * 
 * @param num cislo souboru (0-99)
 * @return char 0, pokud probehlo vse v poradku, nenulova hodnota, pokud doslo k chybe
 */
char get_connections(int num) {
	char n[FILENAME_LEN];
	memset(n, 0, FILENAME_LEN * sizeof(char));
	
	clear_wires();

	n[0] = '/';
	n[1] = num / 10 + '0';
	n[2] = num % 10 + '0';
	n[3] = '.';
	n[4] = 'P';
	n[5] = 'P';
	n[6] = 0;
	

	pin_number pin_buf[INPUT_NR + OUTPUT_NR];
	memset(pin_buf, 0, INPUT_NR + OUTPUT_NR * sizeof(pin_number));
	pin_number pin_temp = 0;
	int i = 0;
	char r;

	unsigned char line = 1;
	unsigned char connection = 1;

	wires_used = 1;

	Serial.printf("Filename: %s\n", n);

	if(num < 0) {
		return 0;
	}
	
	if(!SPIFFS.exists(n)) {
		return 1;
	}
	
	File file = SPIFFS.open(n, "r");

	if (!file) {
		Serial.println("Open ERR");
		return 1;
	}

	
	while(file.available()) {
		r = file.read();
		
		//odradkovani muze prijit jen mezi propojkami
		if(r == '\n') {
			++line;
			connection = 0;
			continue;
		}

		i = 0;
		//nacitam vzdy jednu propojku naraz (jsou oddelene stredniky)
		while(1) {
			if(r == '-' || r == ';') {
				Serial.printf("Nacten pin %d\n", (int)pin_temp);
				pin_buf[i] = pin_temp;
				++i;
				pin_temp = 0;
				
				if(r == ';') {
					break;
				}
			}
			else {
				//ocekavam jen cisla, jinak chyba
				if(r >= '0' && r <= '9') {
					pin_temp = pin_temp * 10 + r - '0';
				}
				else {
					show_error(PRE_CONN_ERR, line, connection);
					file.close();
					return 1;
				}
			}
			//v tomto miste musi byt jeste nejake znaky ke cteni
			if(!file.available()) {
				show_error(PRE_CONN_ERR, line, connection);
				file.close();
				return 1;
			}
			r = file.read();
		}

		//mam nacteny jeden propoj, vytvorim z nej drat
		if(add_wire(pin_buf)) {
			//pokud add_wire vratil nenulovou hodnotu, doslo k chybe
			show_error(INV_CONN_ERR, 0, pin_buf[0]);
			file.close();
			return 1;
		}
		memset(pin_buf, 0, INPUT_NR + OUTPUT_NR);
		++connection;
	}
  	file.close();

	//"zazalohuju" stav propojek po nacteni souboru, abych se k nemu mohl rychle vracet
	memcpy(default_wire_inputs, wire_inputs, WIRE_AMOUNT * INPUT_NR);
	memcpy(default_wire_outputs, wire_outputs, WIRE_AMOUNT * OUTPUT_NR);
	
	memcpy(default_pins_wires, pins_wires, PIN_AMOUNT);
  
  	return 0;
}

void run_blocks() {
	//pokud pri pomale simulaci jeste neubehlo dost casu od posledni aktualizace
	if(slow_sim_enable == TIME_ON && slow_last_run + SLOW_SIM_DELAY > millis()) {
		return;
	}
	for(int i = 0; i < BLOCK_AMOUNT; ++i) {
		++sim_time;
		if(!compute_order[i]) { //nulove cislo bloku znamena konec sekvence
			break;
		}
		#ifdef PRINT
		Serial.print("Volam funkci pro blok ");
		Serial.printf("%d ", compute_order[i]);
		Serial.printf("%p\n", blocks[compute_order[i]].function_pointer);
		#endif

		execute_block(compute_order[i]);
		
		#ifdef PRINT
		delay(1000);
		#endif
	}
	slow_last_run = millis();
}

/**
 * @brief Ziska stav fyzickych propojek komunikaci s podrizenymi procesory na sbernici I2C
 * 
 * @return char 0, pokud vse probehlo v poradku, nenulove cislo pokud jsou propojky chybne
 */
char get_physical_connections() {

	unsigned char result = 0;
	
	//nactu vychozi hodnoty
	memcpy(wire_inputs, default_wire_inputs, WIRE_AMOUNT * INPUT_NR);
	memcpy(wire_outputs, default_wire_outputs, WIRE_AMOUNT * OUTPUT_NR);	
	memcpy(pins_wires, default_pins_wires, PIN_AMOUNT);
	
	pin_number pin_buf[INPUT_NR + OUTPUT_NR];
	pin_number pin = 0;
	unsigned char index = 0;
	char r;
	unsigned char bitmap[AUX_PORT_AMOUNT * AUX_PORT_WIDTH];
	unsigned char j = 0;
	unsigned long time = 0;
	unsigned long timeout = 100;

	memset(bitmap, 0, AUX_PORT_AMOUNT *AUX_PORT_WIDTH);
	
	//potrebuji projit vsechny radky matice
	for(int row = 0; row < AUX_PORT_AMOUNT * AUX_PORT_WIDTH * aux_devices; ++row) {
		//v kazdem radku potrebuji po bytech pres I2C nacist stav
		
		//pokud uz je pin v bitmape, uz jsem ho nacetl v ramci predchozich dratu, mohu ho preskocit
		//"zapis" do slave procesory interpretuji jako posun aktivniho pinu
		if(bitmap_is_set(bitmap, row)) {
			Wire.beginTransmission(SLAVE_ADDR);
			//posilana data mohou byt libovolna
			Wire.write(SLAVE_ADDR);
			Wire.endTransmission();

			#ifdef PRINT
			Serial.printf("Preskakuji pin %d.\n", row);
			#endif

			continue;
		}

		//zadost o prvni data radky
		Wire.requestFrom(SLAVE_ADDR, 1);

		//v kazde radce musim nacist stavy registru po 8 bitech
		//radky nacitam hned od registru, kde je aktivni pin
		for(int i = row / 8; i < (AUX_PORT_AMOUNT * AUX_PORT_WIDTH * aux_devices) / 8; ++i) {
			
			run_blocks();
			
			#ifdef PRINT
			//Serial.printf("Ocekavam data z periferie.\n");
			#endif

			time = millis();
			while(!Wire.available()) {
				//explicitni timeout, kdyby doslo k chybe
				if(millis() - time > timeout) {
					return 1;
				}
			}
			//piny jsou realne na pullup rezistorech, spojene maji nizkou uroven
			r = ~Wire.read();

			if(i != (AUX_PORT_AMOUNT * AUX_PORT_WIDTH * aux_devices) / 8 - 1) {
				//pokud nejsem v poslednim kole nacitani radky, zazadam o dalsi data
				Wire.requestFrom(SLAVE_ADDR, 1);
			}

			#ifdef PRINT
			Serial.printf("Z I2C prijato: %d\n", (unsigned char)r);
			delay(2000);
			#endif

			j = 0;
			while(r) {
				if(r & (1 << j)) {
					pin = i * 8 + j;
					pin_buf[index] = pin + 1; //pin 0 se interne nevyuziva, fyzicke piny zacinaji od 1
					++index;
					
					//index co jsem zpracoval znuluju, muze to urychlit pruchod cyklem (leze se do nej pouze pokud je nenulove r)
					r = r & ~(1 << j);
					
					//do bitmapy zanesu vsechny piny mnoziny jako pouzite, v dalsich pruchodech je muzu preskocit (usetrit nacitani jedne cele radky)
					bitmap_set(bitmap, pin);
					#ifdef PRINT
					Serial.printf("Zapisuji do bitmapy pin %d.\n", pin);
					#endif		
				}
				++j;
			}
		}
			
		if(index < 2) {
			//pokud se v mnozine nachazi jen jeden pin, tak neni nikam pripojeny
			memset(pin_buf, 0, INPUT_NR + OUTPUT_NR);
			index = 0;
			continue;
		}
		if(add_wire(pin_buf)) {
			show_error(PHY_CONN_ERR, 0, pin_buf[0]);
			++result;
			//pri chybe zapojeni fyzickych pinu nemuzu funkci ukoncit, musim donacist i pripadny zbytek, aby procesory nezustaly v nekonzistentnim stavu
		}
		memset(pin_buf, 0, 10);
		index = 0;

		#ifdef PRINT
		Serial.printf("\n");
		#endif
	}

	#ifdef PRINT
	Serial.printf("Dokonceno nacitani fyzickych propojek.\n");
	#endif
	
	return result;
}


/**
 * @brief Zjisti jestli nedoslo ke stisku tlacitka u displeje a pripadne upravi cisla konfiguracnich souboru
 * 
 * @return unsigned char 0 pokud nedoslo ke stisku, 1 pokud se zmenily propojky, 2 pokud se zmenily soucastky
 */
unsigned char button_check() {
	unsigned char buttons = disp.readButtons();
	unsigned char changed = buttons ^ last_buttons;
	unsigned char result = 0;

	last_buttons = buttons;
  
	if(millis() - disp_button_debounce < DEBOUNCE_TIME) {
		return result;
	}

	if(changed & BUT8 && buttons & BUT8) { //stisk S8
		if(connection_conf_nr < 99) { //zvysim cislo konfigurace, pokud uz neni na maximu
			++connection_conf_nr;
			result = 1;
		}
	}
	if(changed & BUT7 && buttons & BUT7) { //stisk S7
		if(connection_conf_nr > -1) { //snizim cislo konfigurace, pokud uz neni na minimu
			--connection_conf_nr;
			result = 1;
		}
	}
	if(changed & BUT6 && buttons & BUT6) { //stisk S6
		if(device_conf_nr < 99) { //zvysim cislo konfigurace, pokud uz neni na maximu
			++device_conf_nr;
			result = 2;
		}
	}
	if(changed & BUT5 && buttons & BUT5) { //stisk S5
		if(device_conf_nr > 0) { //snizim cislo konfigurace, pokud uz neni na minimu
			--device_conf_nr;
			result = 2;
		}
	}
	if(changed & BUT2 && buttons & BUT2) { //stisk S2
		if(slow_sim_enable == TIME_ON) { //toggle "pomale" simulace
			slow_sim_enable = TIME_OFF;
		}
		else {
			slow_sim_enable = TIME_ON;
		}
	}
	if(changed & BUT1 && buttons & BUT1) { //stisk S1
		if(sim_time_enable == TIME_ON) { //toggle simulace zpozdeni bloku
			sim_time_enable = TIME_OFF;
		}
		else {
			sim_time_enable = TIME_ON;
		}
	}
	return result;
}

/**
 * @brief Inicializuje mikrokontrolery zajistujici IO expand, tj. zjisti jejich pocet a nastavi jim poradi
 * 
 */
void aux_init() {
	char amount;
	unsigned long timeout = 100;
	unsigned long time;
	digitalWrite(TOKEN_PIN, HIGH);

	for(amount = 0; amount < AUX_MAX_DEV; ++amount) {
		//zkusim, jestli neco vubec posloucha
		Wire.requestFrom(SLAVE_ADDR, 1);
		time = millis();
		while(!Wire.available()) {
			//pokud se do timeoutu nic neozve, tak neni pripojeno vubec nic
			if(millis() - time > timeout) {
				#ifdef PRINT
				Serial.printf("Nepripojen zadny obvod\n");
				#endif
				aux_devices = 0;
				return;
			}
		}
		//pokud se nejaky procesor ozval, cyklus bezi dal
		//prijata data se zahodi, nejsou podstatna
		//pokud se vratil EMPTY_CHAR, tak nikdo nema povereni/vsichni uz jsou inicializovani
		if(Wire.read() == EMPTY_CHAR) {
			#ifdef PRINT
			Serial.printf("Detekovano %d pripojenych obvodu.\n", amount);
			#endif
			
			aux_devices = amount;
			//sam odeslu ukoncovaci znak, init konci
			Wire.beginTransmission(SLAVE_ADDR);
			Wire.write(END_CHAR);
			Serial.printf("End error: %d\n", Wire.endTransmission());
			return;
		}
		#ifdef PRINT
		Serial.printf("Prijato neco jineho nez EMPTY_CHAR.\n");
		#endif
		Wire.beginTransmission(SLAVE_ADDR);
		//na sbernici se odesle index, jaky si ma aktivni procesor prevzit (podle toho jestli ma povereni)
		Wire.write((uint8_t)amount);
		Serial.printf("Error %d\n", Wire.endTransmission());
		delay(2000);
	}
}

/**
 * @brief Nastavi piny pro ucely fyzickych vystupu ze stavebnice
 * 
 */
void pin_setup() {
	pinMode(BUT0_PIN, INPUT_PULLUP);
	pinMode(BUT1_PIN, INPUT_PULLUP);

	pinMode(FOTOR0_PIN, INPUT);
	pinMode(FOTOR1_PIN, INPUT);

	pinMode(IR_PIN, INPUT);

	pinMode(TOKEN_PIN, OUTPUT);
	digitalWrite(TOKEN_PIN, LOW);

	ledcSetup(REPRO_PWM_CHANNEL, MOTOR_PWM_FREQ, MOTOR_BIT_DEPTH);
	ledcAttachPin(REPRO_PIN, REPRO_PWM_CHANNEL);
	ledcWrite(REPRO_PWM_CHANNEL, 0);

	ledcSetup(SERVO_PWM_CHANNEL, SERVO_FREQ, SERVO_BIT_DEPTH);
	ledcAttachPin(SERVO_PIN, SERVO_PWM_CHANNEL);
	ledcWrite(SERVO_PWM_CHANNEL, SERVO_RIGHT);

	ledcSetup(MOTOR_PWM_CHANNEL, MOTOR_PWM_FREQ, MOTOR_BIT_DEPTH);
	ledcAttachPin(MOTOR_PIN, MOTOR_PWM_CHANNEL);
	ledcWrite(MOTOR_PWM_CHANNEL, MOTOR_OFF);

	pinMode(LED0_PIN, OUTPUT);
	digitalWrite(LED0_PIN, LOW);
	pinMode(LED1_PIN, OUTPUT);
	digitalWrite(LED1_PIN, LOW);
	pinMode(LED2_PIN, OUTPUT);
	digitalWrite(LED2_PIN, LOW);
	pinMode(LED3_PIN, OUTPUT);
	digitalWrite(LED3_PIN, LOW);
}


void setup(void) {
	Serial.begin(115200);
	WiFi.mode(WIFI_STA);
	WiFi.begin(ssid, password);
	Serial.println("");

	// Wait for connection
	while (WiFi.status() != WL_CONNECTED) {
		delay(500);
		Serial.print(".");
	}
	Serial.println("");
	Serial.print("Connected to ");
	Serial.println(ssid);
	Serial.print("IP address: ");
	Serial.println(WiFi.localIP());

	server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
		request->send(200, "text/plain", "Hi! I am ESP32.");
	});

	AsyncElegantOTA.begin(&server);    // Start ElegantOTA
	server.begin();
	Serial.println("HTTP server started");
	
	if (!SPIFFS.begin(true)) {
		Serial.println("An Error has occurred while mounting SPIFFS");
		return;
	}

	disp.displayBegin();
	display_refresh();

	memset(wire_inputs, 0, WIRE_AMOUNT * sizeof(pin_number));
	memset(wire_outputs, 0, WIRE_AMOUNT * sizeof(pin_number));
  	//"rozkouskovani" pole vstupu a vystupu do konkretnich dratu
	for(int i = 0; i < WIRE_AMOUNT; ++i) {
		wires[i].inputs = wire_inputs + i * INPUT_NR;
		wires[i].outputs = wire_outputs + i * OUTPUT_NR;
	}

	memset(pins_wires, 0, PIN_AMOUNT * sizeof(wire_number));
	//"rozkouskovani" pole cisel dratu, ke kterym jsou pripojene piny
	for(int i = 0; i < PIN_AMOUNT; ++i) {
		pins[i].wire_nr = pins_wires + i;
	}

	Wire.begin(SDA_PIN, SCL_PIN, I2C_SPEED);

	pin_setup();

	aux_init();

	//nulovy drat znamena "nepripojeno nikam"
	wires[0].wire_state = STATE_UNDEF;
	get_devices(0);
}

void loop(void) {
	
	char result;

	result = button_check();
	if(result) {
		display_refresh();
		disp_err = NO_ERR;

		if(result == 2) {
			get_devices(device_conf_nr);
		}
		//propojky se musi nacist, i pokud byly zmeneny jenom soucastky
		//nebudou se nacitat pokud doslo k chybe pri nacitani soucastek nebo pokud byly vypnuty
		if(disp_err == NO_ERR && connection_conf_nr >= 0) {
			get_connections(connection_conf_nr);
		}
		//pokud zustal stav bez erroru, tak displej zhasne
		display_refresh();
	}
	
	//vypocet bloku bezi tehdy, kdyz bud error neni, nebo je runtimovy (zkratovani vystupu), nebo ho zpusobily fyzicke propojky
	if(disp_err == NO_ERR || disp_err == RUN_ERR || disp_err == PHY_CONN_ERR) {
		
		if(disp_err == PHY_CONN_ERR || millis() - last_aux_update > AUX_UPDATE_INTERVAL) {
			disp_err = 0;
			display_refresh();
			get_physical_connections();
			get_block_order();
			last_aux_update = millis();
		}	

		//chybu v tomto miste mohly zpusobit jen fyzicke propojky
		if(disp_err != PHY_CONN_ERR) {
			disp_err = NO_ERR;

			run_blocks();

			//pokud se vsechny bloky zpracovaly OK, tak disp_err zustal NO_ERR a chyba na displeji zhasne
			display_refresh();
		}
	}
}
